from __future__ import print_function

import itertools
import json
import math
from math import ceil
from pathlib import Path
from copy import deepcopy

import gym
import numpy as np
import pygame
import scipy.stats
from gym.spaces import Box, Discrete
from scipy.spatial import ConvexHull
from scipy.stats import norm

EPS_SPEED = 0.001  # Minimum speed to be considered in motion
HOME_X = .0
HOME_Y = .0
HOME_POS = (HOME_X, HOME_Y)

ROTATION = 10
MAX_SPEED = 1.

FUEL_MINE = -.05
FUEL_ACC = -.025
FUEL_IDLE = -0.005

CAPACITY = 1

ACT_MINE = 0
ACT_LEFT = 1
ACT_RIGHT = 2
ACT_ACCEL = 3
ACT_BRAKE = 4
ACT_NONE = 5
FUEL_LIST = [FUEL_MINE + FUEL_IDLE, FUEL_IDLE, FUEL_IDLE,
             FUEL_IDLE + FUEL_ACC, FUEL_IDLE, FUEL_IDLE]
FUEL_DICT = {ACT_MINE: FUEL_MINE + FUEL_IDLE, ACT_LEFT: FUEL_IDLE, ACT_RIGHT: FUEL_IDLE,
             ACT_ACCEL: FUEL_IDLE + FUEL_ACC, ACT_BRAKE: FUEL_IDLE, ACT_NONE: FUEL_IDLE}
ACTIONS = ["Mine", "Left", "Right", "Accelerate", "Brake", "None"]
ACTION_COUNT = len(ACTIONS)


MINE_RADIUS = 0.14
BASE_RADIUS = 0.15

WIDTH = 480
HEIGHT = 480

# Color definitions
WHITE = (255, 255, 255)
GRAY = (150, 150, 150)
C_GRAY = (150 / 255., 150 / 255., 150 / 255.)
DARK_GRAY = (100, 100, 100)
BLACK = (0, 0, 0)
RED = (255, 70, 70)
C_RED = (1., 70 / 255., 70 / 255.)

FPS = 180

MINE_LOCATION_TRIES = 100

MINE_SCALE = 1.
BASE_SCALE = 1.
CART_SCALE = 1.

MARGIN = 0.16 * CART_SCALE

ACCELERATION = 0.0075 * CART_SCALE
DECELERATION = 1

CART_IMG = str(Path(__file__).parent.absolute()) + '/assets/cart.png'
MINE_IMG = str(Path(__file__).parent.absolute()) + '/assets/mine.png'


class Mine:
    """Class representing an individual Mine
    """

    def __init__(self, ore_cnt, x, y):
        self.distributions = [
            scipy.stats.norm(np.random.random(), np.random.random())
            for _ in range(ore_cnt)
        ]
        self.pos = np.array((x, y))

    def distance(self, cart):
        return mag(cart.pos - self.pos)

    def mineable(self, cart):
        return self.distance(cart) <= MINE_RADIUS * MINE_SCALE * CART_SCALE

    def mine(self):
        """Generates collected resources according to the mine's random
        distribution

        Returns:
            list -- list of collected ressources
        """
        return [max(0., dist.rvs()) for dist in self.distributions]

    def distribution_means(self):
        """
            Computes the mean of the truncated normal distributions
        """
        means = np.zeros(len(self.distributions))

        for i, dist in enumerate(self.distributions):
            mean, std = dist.mean(), dist.std()
            means[i] = truncated_mean(mean, std, 0, float("inf"))
            if np.isnan(means[i]):
                means[i] = 0
        return means


class Cart:
    """Class representing the actual minecart
    """
    def __init__(self, ore_cnt):
        self.ore_cnt = ore_cnt
        self.pos = np.array([HOME_X, HOME_Y])
        self.speed = 0
        self.angle = 45
        self.content = np.zeros(self.ore_cnt)
        self.departed = False  # Keep track of whether the agent has left the base

    def accelerate(self, acceleration):
        self.speed = clip(self.speed + acceleration, 0, MAX_SPEED)

    def rotate(self, rotation):
        self.angle = (self.angle + rotation) % 360

    def step(self):
        """
            Update cart's position, taking the current speed into account
            Colliding with a border at anything but a straight angle will cause
            cart to "slide" along the wall.
        """
        pre = np.copy(self.pos)
        if self.speed < EPS_SPEED:
            return False
        x_velocity = self.speed * math.cos(self.angle * math.pi / 180)
        y_velocity = self.speed * math.sin(self.angle * math.pi / 180)
        x, y = self.pos
        if y != 0 and y != 1 and (y_velocity > 0 + EPS_SPEED or
                                  y_velocity < 0 - EPS_SPEED):
            if x == 1 and x_velocity > 0:
                self.angle += math.copysign(ROTATION, y_velocity)
            if x == 0 and x_velocity < 0:
                self.angle -= math.copysign(ROTATION, y_velocity)
        if x != 0 and x != 1 and (x_velocity > 0 + EPS_SPEED or
                                  x_velocity < 0 - EPS_SPEED):
            if y == 1 and y_velocity > 0:
                self.angle -= math.copysign(ROTATION, x_velocity)

            if y == 0 and y_velocity < 0:
                self.angle += math.copysign(ROTATION, x_velocity)

        self.pos[0] = clip(x + x_velocity, 0, 1)
        self.pos[1] = clip(y + y_velocity, 0, 1)
        self.speed = mag(pre - self.pos)

        return True


class Minecart(gym.Env):
    """Minecart environment
    """
    metadata = {'render_modes': ['rgb_array', 'human'], 'render_fps': FPS}

    def __init__(self,
                 image_observation=False,
                 config=str(Path(__file__).parent.absolute()) + '/mine_config.json'):

        self.screen = None
        self.last_render_mode_used = None
        self.config = config

        with open(self.config) as f:
            data = json.load(f)
        
        self.ore_cnt = data['ore_cnt']
        self.capacity = data['capacity']
        self.mine_cnt = data['mine_cnt']
        ore_colors = None if "ore_colors" not in data else data["ore_colors"]
        self.ore_colors = ore_colors or [(np.random.randint(40, 255), np.random.randint(40, 255), np.random.randint(40, 255)) for i in range(self.ore_cnt)]
        self.generate_mines(None)

        if "mines" in data:
            for mine_data, mine in zip(data["mines"], self.mines):
                mine.pos = np.array([mine_data["x"], mine_data["y"]])
                if "distributions" in mine_data:
                    mine.distributions = [
                        scipy.stats.norm(dist[0], dist[1])
                        for dist in mine_data["distributions"]
                    ]

        self.cart = Cart(self.ore_cnt)

        self.end = False

        self.image_observation = image_observation
        if self.image_observation:
            shape = (WIDTH, HEIGHT, 3)
            self.observation_space = Box(low=np.zeros(shape), high=255*np.ones((WIDTH, HEIGHT, 3)), dtype=np.uint8)
        else:
            self.observation_space = Box(np.zeros(7), np.ones(7), dtype=np.float32)
        
        self.action_space = Discrete(6)
        self.reward_space = Box(low=-1, high=self.capacity, shape=(self.ore_cnt + 1,))

    def obj_cnt(self):
        return self.ore_cnt + 1

    def convex_coverage_set(self, frame_skip=1, discount=0.98, incremental_frame_skip=True, symmetric=True):
        """
            Computes an approximate convex coverage set

            Keyword Arguments:
                frame_skip {int} -- How many times each action is repeated (default: {1})
                discount {float} -- Discount factor to apply to rewards (default: {1})
                incremental_frame_skip {bool} -- Wether actions are repeated incrementally (default: {1})
                symmetric {bool} -- If true, we assume the pattern of accelerations from the base to the mine is the same as from the mine to the base (default: {True})

            Returns:
                The convex coverage set 
        """
        policies = self.pareto_coverage_set(
            frame_skip, discount, incremental_frame_skip, symmetric)
        origin = np.min(policies, axis=0)
        extended_policies = [origin] + policies
        return [policies[idx - 1] for idx in ConvexHull(extended_policies).vertices if idx != 0]

    def pareto_coverage_set(self, frame_skip=1, discount=0.98, incremental_frame_skip=True, symmetric=True):
        """
            Computes an approximate pareto coverage set

            Keyword Arguments:
                frame_skip {int} -- How many times each action is repeated (default: {1})
                discount {float} -- Discount factor to apply to rewards (default: {1})
                incremental_frame_skip {bool} -- Wether actions are repeated incrementally (default: {1})
                symmetric {bool} -- If true, we assume the pattern of accelerations from the base to the mine is the same as from the mine to the base (default: {True})

            Returns:
                The pareto coverage set
        """
        all_rewards = []
        base_perimeter = BASE_RADIUS * BASE_SCALE

        # Empty mine just outside the base
        virtual_mine = Mine(self.ore_cnt, (base_perimeter**2 / 2)
                            ** (1 / 2), (base_perimeter**2 / 2)**(1 / 2))
        virtual_mine.distributions = [
            scipy.stats.norm(0, 0)
            for _ in range(self.ore_cnt)
        ]
        for mine in (self.mines + [virtual_mine]):
            mine_distance = mag(mine.pos - HOME_POS) - \
                MINE_RADIUS * MINE_SCALE - BASE_RADIUS * BASE_SCALE / 2

            # Number of rotations required to face the mine
            angle = compute_angle(mine.pos, HOME_POS, [1, 1])
            rotations = int(ceil(abs(angle) / (ROTATION * frame_skip)))

            # Build pattern of accelerations/nops to reach the mine
            # initialize with single acceleration
            queue = [{"speed": ACCELERATION * frame_skip, "dist": mine_distance - frame_skip *
                      (frame_skip + 1) / 2 * ACCELERATION if incremental_frame_skip else mine_distance - ACCELERATION * frame_skip * frame_skip, "seq": [ACT_ACCEL]}]
            trimmed_sequences = []

            while len(queue) > 0:
                seq = queue.pop()
                # accelerate
                new_speed = seq["speed"] + ACCELERATION * frame_skip
                accelerations = new_speed / ACCELERATION
                movement = accelerations * (accelerations + 1) / 2 * ACCELERATION - (
                    accelerations - frame_skip) * ((accelerations - frame_skip) + 1) / 2 * ACCELERATION
                dist = seq["dist"] - movement
                speed = new_speed
                if dist <= 0:
                    trimmed_sequences.append(seq["seq"] + [ACT_ACCEL])
                else:
                    queue.append({"speed": speed, "dist": dist,
                                  "seq": seq["seq"] + [ACT_ACCEL]})
                # idle
                dist = seq["dist"] - seq["speed"] * frame_skip

                if dist <= 0:
                    trimmed_sequences.append(seq["seq"] + [ACT_NONE])
                else:
                    queue.append(
                        {"speed": seq["speed"], "dist": dist, "seq": seq["seq"] + [ACT_NONE]})

            # Build rational mining sequences
            mine_means = mine.distribution_means() * frame_skip
            mn_sum = np.sum(mine_means)
            # on average it takes up to this many actions to fill cart
            max_mine_actions = 0 if mn_sum == 0 else int(
                ceil(self.capacity / mn_sum))

            # all possible mining sequences (i.e. how many times we mine)
            mine_sequences = [[ACT_MINE] *
                              i for i in range(1, max_mine_actions + 1)]

            # All possible combinations of actions before, during and after mining
            if len(mine_sequences) > 0:
                if not symmetric:
                    all_sequences = map(
                        lambda sequences: list(sequences[0]) + list(sequences[1]) + list(
                            sequences[2]) + list(
                            sequences[3]) + list(
                            sequences[4]),
                        itertools.product([[ACT_LEFT] * rotations],
                                          trimmed_sequences,
                                          [[ACT_BRAKE] + [ACT_LEFT] *
                                              (180 // (ROTATION * frame_skip))],
                                          mine_sequences,
                                          trimmed_sequences)
                    )

                else:
                    all_sequences = map(
                        lambda sequences: list(sequences[0]) + list(sequences[1]) + list(
                            sequences[2]) + list(
                            sequences[3]) + list(
                            sequences[1]),
                        itertools.product([[ACT_LEFT] * rotations],
                                          trimmed_sequences,
                                          [[ACT_BRAKE] + [ACT_LEFT] *
                                              (180 // (ROTATION * frame_skip))],
                                          mine_sequences)
                    )
            else:
                if not symmetric:
                    print([ACT_NONE] + trimmed_sequences[1:],
                          trimmed_sequences[1:], trimmed_sequences)
                    all_sequences = map(
                        lambda sequences: list(sequences[0]) + list(sequences[1]) + list(
                            sequences[2]) + [ACT_NONE] + list(
                            sequences[3])[1:],
                        itertools.product([[ACT_LEFT] * rotations],
                                          trimmed_sequences,
                                          [[ACT_LEFT] *
                                              (180 // (ROTATION * frame_skip))],
                                          trimmed_sequences)
                    )

                else:
                    all_sequences = map(
                        lambda sequences: list(sequences[0]) + list(sequences[1]) + list(
                            sequences[2]) + [ACT_NONE] + list(
                            sequences[1][1:]),
                        itertools.product([[ACT_LEFT] * rotations],
                                          trimmed_sequences,
                                          [[ACT_LEFT] * (180 // (ROTATION * frame_skip))])
                    )

            # Compute rewards for each sequence
            fuel_costs = np.array([f * frame_skip for f in FUEL_LIST])

            def maxlen(l):
                if len(l) == 0:
                    return 0
                return max([len(s) for s in l])

            longest_pattern = maxlen(trimmed_sequences)
            max_len = rotations + longest_pattern + 1 + \
                (180 // (ROTATION * frame_skip)) + \
                maxlen(mine_sequences) + longest_pattern
            discount_map = discount**np.arange(max_len)
            for s in all_sequences:
                reward = np.zeros((len(s), self.obj_cnt()))
                reward[:, -1] = fuel_costs[s]
                mine_actions = s.count(ACT_MINE)
                reward[-1, :-1] = mine_means * mine_actions / \
                    max(1, (mn_sum * mine_actions) / self.capacity)

                reward = np.dot(discount_map[:len(s)], reward)
                all_rewards.append(reward)

            all_rewards = pareto_filter(all_rewards, minimize=False)

        return all_rewards

    def generate_mines(self, mine_distributions=None):
        """
            Randomly generate mines that don't overlap the base
            TODO: propose some default formations
        """
        self.mines = []
        for i in range(self.mine_cnt):
            pos = np.array((np.random.random(), np.random.random()))

            tries = 0
            while (mag(pos - HOME_POS) < BASE_RADIUS * BASE_SCALE + MARGIN) and (tries < MINE_LOCATION_TRIES):
                pos[0] = np.random.random()
                pos[1] = np.random.random()
                tries += 1
            assert tries < MINE_LOCATION_TRIES
            self.mines.append(Mine(self.ore_cnt, *pos))
            if mine_distributions:
                self.mines[i].distributions = mine_distributions[i]

    def initialize_mines(self):
        """Assign a random rotation to each mine, and initialize the necessary sprites
        for the Pygame backend
        """

        for mine in self.mines:
            mine.rotation = np.random.randint(0, 360)

        self.mine_sprites = pygame.sprite.Group()
        self.mine_rects = []
        for mine in self.mines:
            mine_sprite = pygame.sprite.Sprite()
            mine_sprite.image = pygame.transform.rotozoom(
                pygame.image.load(MINE_IMG), mine.rotation,
                MINE_SCALE).convert_alpha()
            self.mine_sprites.add(mine_sprite)
            mine_sprite.rect = mine_sprite.image.get_rect()
            mine_sprite.rect.centerx = (
                mine.pos[0] * (1 - 2 * MARGIN)) * WIDTH + MARGIN * WIDTH
            mine_sprite.rect.centery = (
                mine.pos[1] * (1 - 2 * MARGIN)) * HEIGHT + MARGIN * HEIGHT
            self.mine_rects.append(mine_sprite.rect)

    def step(self, action, frame_skip=4, incremental_frame_skip=True):
        """Perform the given action `frame_skip` times
         ["Mine", "Left", "Right", "Accelerate", "Brake", "None"]
        Arguments:
            action {int} -- Action to perform, ACT_MINE (0), ACT_LEFT (1), ACT_RIGHT (2), ACT_ACCEL (3), ACT_BRAKE (4) or ACT_NONE (5)

        Keyword Arguments:
            frame_skip {int} -- Repeat the action this many times (default: {1})
            incremental_frame_skip {int} -- If True, frame_skip actions are performed in succession, otherwise the repeated actions are performed simultaneously (e.g., 4 accelerations are performed and then the cart moves).

        Returns:
            tuple -- (state, reward, terminal) tuple
        """
        change = False  # Keep track of whether the state has changed

        if action < 0 or action >= ACTION_COUNT:
            action = ACT_NONE

        reward = np.zeros(self.ore_cnt + 1)
        if frame_skip < 1:
            frame_skip = 1

        reward[-1] = FUEL_IDLE * frame_skip

        if action == ACT_ACCEL:
            reward[-1] += FUEL_ACC * frame_skip
        elif action == ACT_MINE:
            reward[-1] += FUEL_MINE * frame_skip

        for _ in range(frame_skip if incremental_frame_skip else 1):

            if action == ACT_LEFT:
                self.cart.rotate(-ROTATION *
                                 (1 if incremental_frame_skip else frame_skip))
                change = True
            elif action == ACT_RIGHT:
                self.cart.rotate(
                    ROTATION * (1 if incremental_frame_skip else frame_skip))
                change = True
            elif action == ACT_ACCEL:
                self.cart.accelerate(
                    ACCELERATION * (1 if incremental_frame_skip else frame_skip))
            elif action == ACT_BRAKE:
                self.cart.accelerate(-DECELERATION *
                                     (1 if incremental_frame_skip else frame_skip))
            elif action == ACT_MINE:
                for _ in range(1 if incremental_frame_skip else frame_skip):
                    change = self.mine() or change

            if self.end:
                break

            for _ in range(1 if incremental_frame_skip else frame_skip):
                change = self.cart.step() or change

            distanceFromBase = mag(self.cart.pos - HOME_POS)
            if distanceFromBase < BASE_RADIUS * BASE_SCALE:
                if self.cart.departed:
                    # Cart left base then came back, ending the episode
                    self.end = True
                    # Sell resources
                    reward[:self.ore_cnt] += self.cart.content
                    self.cart.content = np.zeros(self.ore_cnt)
            else:
                # Cart left base
                self.cart.departed = True

        if change and self.image_observation:
            self.render_pygame()

        return self.get_state(change), reward, self.end, {}

    def mine(self):
        """Perform the MINE action

        Returns:
            bool -- True if something was mined
        """
        if self.cart.speed < EPS_SPEED:
            # Get closest mine
            mine = min(self.mines, key=lambda mine: mine.distance(self.cart))

            if mine.mineable(self.cart):
                cart_free = self.capacity - np.sum(self.cart.content)
                mined = mine.mine()
                total_mined = np.sum(mined)
                if total_mined > cart_free:
                    # Scale mined content to remaining capacity
                    scale = cart_free / total_mined
                    mined = np.array(mined) * scale

                self.cart.content += mined

                if np.sum(mined) > 0:
                    return True
        return False

    def get_pixels(self, update=True):
        """Get the environment's image representation

        Keyword Arguments:
            update {bool} -- Whether to redraw the environment (default: {True})

        Returns:
            np.array -- array of pixels, with shape (width, height, channels)
        """
        if update:
            self.pixels = pygame.surfarray.array3d(self.screen)

        return self.pixels

    def get_state(self, update=True):
        """Returns the environment's state

        Keyword Arguments:
            update {bool} -- Whether to update the representation (default: {True})

        Returns:
            dict -- dict containing the aforementioned elements
        """
        if self.image_observation:
            state = self.get_pixels(update)
        else:
            angle = math.radians(self.cart.angle)
            sina = math.sin(angle)
            cosa = math.cos(angle)
            angle = np.array([sina, cosa], dtype=np.float32)
            state = np.concatenate((self.cart.pos, np.array([self.cart.speed]), angle, self.cart.content/self.capacity), dtype=np.float32)
        return state

        """ return {
            "position": self.cart.pos,
            "speed": self.cart.speed,
            "orientation": self.cart.angle,
            "content": self.cart.content,
            "pixels": self.get_pixels(update)
        } """

    def reset(self, seed=None, return_info=False, **kwargs):
        """Resets the environment to the start state

        Returns:
            [type] -- [description]
        """
        super().reset(seed=seed)
        self.np_random.seed(seed)

        if self.screen is None and self.image_observation:
            self.render(mode='rgb_array') # init pygame

        if self.image_observation:
            self.render_pygame()
 
        self.cart.content = np.zeros(self.ore_cnt)
        self.cart.pos = np.array(HOME_POS)
        self.cart.speed = 0
        self.cart.angle = 45
        self.cart.departed = False
        self.end = False
        return (self.get_state(), {}) if return_info else self.get_state()

    def __str__(self):
        string = "Completed: {} ".format(self.end)
        string += "Departed: {} ".format(self.cart.departed)
        string += "Content: {} ".format(self.cart.content)
        string += "Speed: {} ".format(self.cart.speed)
        string += "Direction: {} ({}) ".format(self.cart.angle, self.cart.angle * math.pi / 180)
        string += "Position: {} ".format(self.cart.pos)
        return string

    def render(self, mode='human'):
        if self.screen is None or self.last_render_mode_used != mode:
            self.last_render_mode_used = mode
            pygame.init()
            self.screen = pygame.display.set_mode((WIDTH, HEIGHT), flags=pygame.HIDDEN if mode=='rgb_array' else 0)
            self.clock = pygame.time.Clock()

            self.initialize_mines()

            self.cart_sprite = pygame.sprite.Sprite()
            self.cart_sprites = pygame.sprite.Group()
            self.cart_sprites.add(self.cart_sprite)
            self.cart_image = pygame.transform.rotozoom(
                pygame.image.load(CART_IMG).convert_alpha(), 0,
                CART_SCALE)

        if not self.image_observation:
            self.render_pygame()  # if the obs is not an image, then step would not have rendered the screen

        if mode == 'human':
            self.clock.tick(FPS)
            pygame.display.update()
        elif mode == 'rgb_array':
            string_image = pygame.image.tostring(self.screen, 'RGB')
            temp_surf = pygame.image.fromstring(string_image, (WIDTH, HEIGHT), 'RGB')
            tmp_arr = pygame.surfarray.array3d(temp_surf)
            return tmp_arr

    def render_pygame(self):
        pygame.event.get()

        self.mine_sprites.update()

        # Clear canvas
        self.screen.fill(GRAY)

        # Draw Home
        pygame.draw.circle(self.screen, RED, (int(WIDTH * HOME_X), int(
            HEIGHT * HOME_Y)), int(WIDTH / 3 * BASE_SCALE))

        # Draw Mines
        self.mine_sprites.draw(self.screen)

        # Draw cart
        self.cart_sprite.image = rot_center(self.cart_image,
                                            -self.cart.angle).copy()

        self.cart_sprite.rect = self.cart_sprite.image.get_rect(
            center=(200, 200))

        self.cart_sprite.rect.centerx = self.cart.pos[0] * (
            1 - 2 * MARGIN) * WIDTH + MARGIN * WIDTH
        self.cart_sprite.rect.centery = self.cart.pos[1] * (
            1 - 2 * MARGIN) * HEIGHT + MARGIN * HEIGHT

        self.cart_sprites.update()

        self.cart_sprites.draw(self.screen)

        # Draw cart content
        width = self.cart_sprite.rect.width / (2 * self.ore_cnt)
        height = self.cart_sprite.rect.height / 3
        content_width = (width + 1) * self.ore_cnt
        offset = (self.cart_sprite.rect.width - content_width) / 2
        for i in range(self.ore_cnt):

            rect_height = height * self.cart.content[i] / self.capacity

            if rect_height >= 1:
                pygame.draw.rect(self.screen, self.ore_colors[i], (
                    self.cart_sprite.rect.left + offset + i * (width + 1),
                    self.cart_sprite.rect.top + offset * 1.5, width,
                    rect_height))

    def close(self):
        if self.screen is not None:
            pygame.display.quit()
            pygame.quit()

    def __deepcopy__(self, memo):
        this_copy = Minecart(self.image_observation, self.config)
        this_copy.cart = deepcopy(self.cart)
        this_copy.mines = deepcopy(self.mines)
        this_copy.end = self.end
        return this_copy

def compute_angle(p0, p1, p2):
    v0 = np.array(p0) - np.array(p1)
    v1 = np.array(p2) - np.array(p1)

    angle = np.math.atan2(np.linalg.det([v0, v1]), np.dot(v0, v1))
    return np.degrees(angle)


def rot_center(image, angle):
    """Rotate an image while preserving its center and size"""
    orig_rect = image.get_rect()
    rot_image = pygame.transform.rotate(image, angle)
    rot_rect = orig_rect.copy()
    rot_rect.center = rot_image.get_rect().center
    rot_image = rot_image.subsurface(rot_rect).copy()
    return rot_image


def mag(vector2d):
    return np.sqrt(np.dot(vector2d, vector2d))


def clip(val, lo, hi):
    return lo if val <= lo else hi if val >= hi else val


def scl(c):
    return (c[0] / 255., c[1] / 255., c[2] / 255.)


def truncated_mean(mean, std, a, b):
    if std == 0:
        return mean
    from scipy.stats import norm
    a = (a - mean) / std
    b = (b - mean) / std
    PHIB = norm.cdf(b)
    PHIA = norm.cdf(a)
    phib = norm.pdf(b)
    phia = norm.pdf(a)

    trunc_mean = (mean + ((phia - phib) / (PHIB - PHIA)) * std)
    return trunc_mean


def pareto_filter(costs, minimize=True):
    """
    Find the pareto-efficient points
    :param costs: An (n_points, n_costs) array
    :param return_mask: True to return a mask
    :return: An array of indices of pareto-efficient points.
        If return_mask is True, this will be an (n_points, ) boolean array
        Otherwise it will be a (n_efficient_points, ) integer array of indices.
    from https://stackoverflow.com/a/40239615
    """
    costs_copy = np.copy(costs) if minimize else -np.copy(costs)
    is_efficient = np.arange(costs_copy.shape[0])
    n_points = costs_copy.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs_copy):
        nondominated_point_mask = np.any(
            costs_copy < costs_copy[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        # Remove dominated points
        is_efficient = is_efficient[nondominated_point_mask]
        costs_copy = costs_copy[nondominated_point_mask]
        next_point_index = np.sum(
            nondominated_point_mask[:next_point_index]) + 1
    return [costs[i] for i in is_efficient]


if __name__ == '__main__':
    env = Minecart(image_observation=True)
    done = False
    env.reset()
    while True:
        env.render(mode='human')
        obs, r, done, info = env.step(env.action_space.sample())
        #print(str(env))
        if done:
            env.reset()



import numpy as np
from gym.envs.mujoco.hopper_v4 import HopperEnv
from gym.spaces import Box


class MOHopperEnv(HopperEnv):
    def __init__(self, **kwargs):
        super(MOHopperEnv, self).__init__(**kwargs)
        self.reward_space = Box(low=-np.inf, high=np.inf, shape=(3,))
    
    def step(self, action):
        x_position_before = self.data.qpos[0]
        self.do_simulation(action, self.frame_skip)
        x_position_after = self.data.qpos[0]
        x_velocity = (x_position_after - x_position_before) / self.dt

        #ctrl_cost = self.control_cost(action)

        #forward_reward = self._forward_reward_weight * x_velocity
        healthy_reward = self.healthy_reward

        #rewards = forward_reward + healthy_reward
        #costs = ctrl_cost

        observation = self._get_obs()
        #reward = rewards - costs
        done = self.done
        
        z = self.data.qpos[1]
        height = 10*(z - self.init_qpos[1])
        energy_cost = np.sum(np.square(action))
        vec_reward = np.array([x_velocity, height, -energy_cost], dtype=np.float32)
        vec_reward += healthy_reward

        info = {
            "x_position": x_position_after,
            "x_velocity": x_velocity,
            "height_reward": height,
            "energy_reward": -energy_cost,
        }

        return observation, vec_reward, done, info



import math

import numpy as np
from gym import spaces
from gym.envs.classic_control.mountain_car import MountainCarEnv


class MOMountainCar(MountainCarEnv):

    def __init__(self, goal_velocity=0):
        super().__init__(goal_velocity)

        self.reward_space = spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32)
    
    def step(self, action: int):
        assert self.action_space.contains(action), f"{action!r} ({type(action)}) invalid"

        position, velocity = self.state
        velocity += (action - 1) * self.force + math.cos(3 * position) * (-self.gravity)
        velocity = np.clip(velocity, -self.max_speed, self.max_speed)
        position += velocity
        position = np.clip(position, self.min_position, self.max_position)
        if position == self.min_position and velocity < 0:
            velocity = 0

        done = bool(position >= self.goal_position and velocity >= self.goal_velocity)
        #reward = -1.0
        reward = np.zeros(3, dtype=np.float32)
        reward[0] = 0.0 if done else -1.0        # time penalty
        reward[1] = 0.0 if action != 0 else -1.0 # reverse penalty
        reward[2] = 0.0 if action != 2 else -1.0 # forward penalty

        self.state = (position, velocity)
        return np.array(self.state, dtype=np.float32), reward, done, {}



import random

import gym
import numpy as np
import pygame
from gym.spaces import Box, Discrete

MAZE=np.array([
    ['1', ' ', ' ', ' ', ' ', '2', 'X', ' ', ' ', ' ', ' ', ' ', 'G'],
    [' ', ' ', ' ', ' ', ' ', ' ', 'X', ' ', ' ', ' ', ' ', ' ', ' '],
    [' ', ' ', ' ', ' ', ' ', ' ', '1', ' ', ' ', ' ', ' ', ' ', ' '],
    [' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
    [' ', ' ', ' ', ' ', ' ', ' ', 'X', ' ', ' ', ' ', ' ', ' ', ' '],
    ['2', ' ', ' ', ' ', ' ', '3', 'X', ' ', ' ', ' ', ' ', ' ', ' '],
    ['X', 'X', '3', ' ', 'X', 'X', 'X', 'X', 'X', ' ', '1', 'X', 'X'],
    [' ', ' ', ' ', ' ', ' ', ' ', 'X', '2', ' ', ' ', ' ', ' ', '3'],
    [' ', ' ', ' ', ' ', ' ', ' ', 'X', ' ', ' ', ' ', ' ', ' ', ' '],
    [' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
    [' ', ' ', ' ', ' ', ' ', ' ', '2', ' ', ' ', ' ', ' ', ' ', ' '],
    [' ', ' ', ' ', ' ', ' ', ' ', 'X', ' ', ' ', ' ', ' ', ' ', ' '],
    ['_', ' ', ' ', ' ', ' ', ' ', 'X', '3', ' ', ' ', ' ', ' ', '1']])
BLUE = (0, 0, 255)
RED = (255, 0, 0)
GREEN = (0, 128, 0)
BLACK = (0, 0, 0)

class FourRoom(gym.Env):
    """
    A discretized version of the gridworld environment introduced in [1]. Here, an agent learns to 
    collect shapes with positive reward, while avoid those with negative reward, and then travel to a fixed goal.
    The gridworld is split into four rooms separated by walls with passage-ways.
    
    # Code adaptaed from: https://github.com/mike-gimelfarb/deep-successor-features-for-transfer/blob/main/source/tasks/gridworld.py

    References
    ----------
    [1] Barreto, André, et al. "Successor Features for Transfer in Reinforcement Learning." NIPS. 2017.
    """

    LEFT, UP, RIGHT, DOWN = 0, 1, 2, 3

    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 4}
 
    def __init__(self, maze=MAZE):
        """
        Creates a new instance of the shapes environment.
        
        Parameters
        ----------
        maze : np.ndarray
            an array of string values representing the type of each cell in the environment:
                G indicates a goal state (terminal state)
                _ indicates an initial state (there can be multiple, and one is selected at random
                    at the start of each episode)
                X indicates a barrier 
                0, 1, .... 9 indicates the type of shape to be placed in the corresponding cell
                entries containing other characters are treated as regular empty cells
        """
        self.window_size = 512
        self.window = None
        self.clock = None

        self.height, self.width = maze.shape
        self.maze = maze
        shape_types = ['1', '2', '3']
        self.all_shapes = dict(zip(shape_types, range(len(shape_types))))
        
        self.goal = None
        self.initial = []
        self.occupied = set()
        self.shape_ids = dict()
        for c in range(self.width):
            for r in range(self.height):
                if maze[r, c] == 'G':
                    self.goal = (r, c)
                elif maze[r, c] == '_':
                    self.initial.append((r, c))
                elif maze[r, c] == 'X':
                    self.occupied.add((r, c))
                elif maze[r, c] in {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9'}:
                    self.shape_ids[(r, c)] = len(self.shape_ids)
        
        self.action_space = Discrete(4)
        self.observation_space = Box(low=np.zeros(2+len(self.shape_ids)), high=len(self.maze)*np.ones(2+len(self.shape_ids)), dtype=np.int32)
        self.reward_space = Box(low=0, high=1, shape=(3,))

    def state_to_array(self, state):
        s = [element for tupl in state for element in tupl]
        return np.array(s, dtype=np.int32)

    def reset(self, seed=None, return_info=False, **kwargs):
        super().reset(seed=seed)
        self.np_random.seed(seed)

        self.state = (random.choice(self.initial), tuple(0 for _ in range(len(self.shape_ids))))
        return (self.state_to_array(self.state), {}) if return_info else self.state_to_array(self.state)
    
    def step(self, action): 
        old_state = self.state
        (row, col), collected = self.state
        
        # perform the movement
        if action == FourRoom.LEFT: 
            col -= 1
        elif action == FourRoom.UP: 
            row -= 1
        elif action == FourRoom.RIGHT: 
            col += 1
        elif action == FourRoom.DOWN: 
            row += 1
        else:
            raise Exception('bad action {}'.format(action))
        
        # out of bounds, cannot move
        if col < 0 or col >= self.width or row < 0 or row >= self.height:
            return self.state_to_array(self.state), np.zeros(len(self.all_shapes), dtype=np.float32), False, {}

        # into a blocked cell, cannot move
        s1 = (row, col)
        if s1 in self.occupied:
            return self.state_to_array(self.state), np.zeros(len(self.all_shapes), dtype=np.float32), False, {}
        
        # can now move
        self.state = (s1, collected)
        
        # into a goal cell
        if s1 == self.goal:
            phi = np.ones(len(self.all_shapes), dtype=np.float32)
            return self.state_to_array(self.state), phi, True, {}
        
        # into a shape cell
        if s1 in self.shape_ids:
            shape_id = self.shape_ids[s1]
            if collected[shape_id] == 1:
                # already collected this flag
                return self.state_to_array(self.state), np.zeros(len(self.all_shapes), dtype=np.float32), False, {}
            else:
                # collect the new flag
                collected = list(collected)
                collected[shape_id] = 1
                collected = tuple(collected)
                self.state = (s1, collected)
                phi = self.features(old_state, action, self.state)
                return self.state_to_array(self.state), phi, False, {}
        
        # into an empty cell
        return self.state_to_array(self.state), np.zeros(len(self.all_shapes), dtype=np.float32), False, {}

    def features(self, state, action, next_state):
        s1, _ = next_state
        _, collected = state
        nc = len(self.all_shapes)
        phi = np.zeros(nc, dtype=np.float32)
        if s1 in self.shape_ids:
            if collected[self.shape_ids[s1]] != 1:
                y, x = s1
                shape_index = self.all_shapes[self.maze[y, x]]
                phi[shape_index] = 1.
        elif s1 == self.goal:
            phi[nc] = np.ones(nc, dtype=np.float32)
        return phi

    def render(self, mode='human'):
        # The size of a single grid square in pixels
        pix_square_size = self.window_size / 13

        if self.window is None and mode == "human":
            pygame.init()
            pygame.display.init()
            self.window = pygame.display.set_mode((self.window_size, self.window_size))
        if self.clock is None and mode == "human":
            self.clock = pygame.time.Clock()

        canvas = pygame.Surface((self.window_size, self.window_size))
        canvas.fill((255, 255, 255))

        pygame.font.init()
        self.font = pygame.font.SysFont(None, 48)
        img = self.font.render('G', True, BLACK)
        canvas.blit(img, (np.array(self.goal)[::-1]+0.15) * pix_square_size)

        for i in range(self.maze.shape[0]):
            for j in range(self.maze.shape[1]):
                (row, col), collected = self.state
                shape_id = self.shape_ids.get((i, j), 0)
                if collected[shape_id] == 1 and self.maze[i,j] != 'X':
                    continue

                pos = np.array([j, i])
                if self.maze[i,j] == '1':
                    pygame.draw.rect(canvas, BLUE,
                        pygame.Rect(
                            pix_square_size * pos,
                            (pix_square_size, pix_square_size),
                        ),
                    )
                elif self.maze[i,j] == 'X':
                    pygame.draw.rect(canvas, BLACK,
                        pygame.Rect(
                            pix_square_size * pos + 1,
                            (pix_square_size, pix_square_size),
                        ),
                    )
                elif self.maze[i,j] == '2':
                    pygame.draw.polygon(canvas, GREEN, [(pos+np.array([0.5,0.0]))*pix_square_size, (pos+np.array([0.,1.]))*pix_square_size, (pos+1.)*pix_square_size])
                elif self.maze[i,j] == '3':
                    pygame.draw.circle(canvas, RED,
                        (pos + 0.5) * pix_square_size,
                        pix_square_size / 2,
                    )
        
        pygame.draw.circle(
            canvas,
            (125, 125, 125),
            (np.array(self.state[0])[::-1] + 0.5) * pix_square_size,
            pix_square_size / 3,
        )
       
        for x in range(13 + 1):
            pygame.draw.line(
                canvas,
                0,
                (0, pix_square_size * x),
                (self.window_size, pix_square_size * x),
                width=1,
            )
            pygame.draw.line(
                canvas,
                0,
                (pix_square_size * x, 0),
                (pix_square_size * x, self.window_size),
                width=1,
            )

        if mode == "human":
            # The following line copies our drawings from `canvas` to the visible window
            self.window.blit(canvas, canvas.get_rect())
            pygame.event.pump()
            pygame.display.update()

            # We need to ensure that human-rendering occurs at the predefined framerate.
            # The following line will automatically add a delay to keep the framerate stable.
            self.clock.tick(self.metadata["render_fps"])
        else:  # rgb_array
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(canvas)), axes=(1, 0, 2)
            )

    def close(self):
        if self.window is not None:
            pygame.display.quit()
            pygame.quit()


if __name__ == '__main__':

    env = FourRoom()
    done = False
    env.reset()
    while not done:
        env.render()
        obs, r, done, info = env.step(env.action_space.sample())



import numpy as np
from gym.envs.mujoco.half_cheetah_v4 import HalfCheetahEnv
from gym.spaces import Box


class MOHalfCheehtahEnv(HalfCheetahEnv):
    def __init__(self, **kwargs):
        super(MOHalfCheehtahEnv, self).__init__(**kwargs)
        self.reward_space = Box(low=-np.inf, high=np.inf, shape=(2,))
    
    def step(self, action):
        observation, reward, done, info = super().step(action)
        vec_reward = np.array([info['reward_run'], info['reward_ctrl']], dtype=np.float32)
        return observation, vec_reward, done, info



from gym import Env
from gym.spaces import Dict, Discrete, MultiBinary, Box
import numpy as np


class BreakableBottles(Env):
    metadata = {"render_modes": ["human", "rgb_array"]}

    # actions
    LEFT = 0
    RIGHT = 1
    PICKUP = 2

    def __init__(self, size=5, prob_drop=0.1, time_penalty=-1, bottle_reward=25, unbreakable_bottles=False, seed=None):
        # settings
        self.prob_drop = prob_drop
        self.time_penalty = time_penalty
        self.bottle_reward = bottle_reward
        self.unbreakable_bottles = False 

        # properties
        self.num_objectives = 3

        # initialize env state
        self.size = size
        self.location = 0
        self.bottles_carrying = 0
        self.bottles_delivered = 0
        self.bottles_dropped = [0]*(self.size - 2)

        # observation and action space
        self.observation_space = Dict(
            {
                "location": Discrete(self.size),
                "bottles_carrying": Discrete(3),
                "bottles_delivered": Discrete(2),
                "bottles_dropped": MultiBinary(self.size - 2),
            }
        )
        self.num_observations = 240

        self.action_space = Discrete(3) # LEFT, RIGHT, PICKUP
        self.num_actions = 3

        # reward space
        self.reward_space = Box(np.array([-np.inf, 0, -1]), np.array([0, self.bottle_reward*2, 0]))

    def step(self, action):
        observation_old = self._get_obs()
        old_potential = self.potential(observation_old)
        terminal = False
        reward = [self.time_penalty, 0, 0]

        if action == self.LEFT and self.location > 0:
            # execute bottle drop, if agent is carrying at least two and current location is 1, 2 or 3
            if self.location in range(1, self.size - 1) and self.bottles_carrying > 1 and np.random.random() < self.prob_drop:
                self.bottles_carrying -= 1
                self.bottles_dropped[self.location - 1] += 1

            # move to the left
            self.location -= 1

        elif action == self.RIGHT and self.location < self.size - 1:
            # execute bottle drop, if agent is carrying at least two and current location is 1, 2 or 3
            if self.location in range(1, self.size - 1) and self.bottles_carrying > 1 and np.random.random() < self.prob_drop:
                self.bottles_carrying -= 1
                self.bottles_dropped[self.location - 1] += 1

            # move to the right
            self.location += 1

            # if agent enters destination tile, deliver any carried bottles
            if self.location == self.size - 1 and self.bottles_carrying > 0:
                num_before = self.bottles_delivered
                num_after = min(self.bottles_delivered + self.bottles_carrying, 2)
                num_delivered = num_after - num_before
                self.bottles_delivered = num_after
                self.bottles_carrying = 0
                reward[1] += self.bottle_reward*num_delivered
                if self.bottles_delivered == 2:
                    terminal = True

        elif action == self.PICKUP:
            # agent is at source and has not reached carrying limit
            if self.location == 0 and self.bottles_carrying < 2:
                # add bottle to inventory
                self.bottles_carrying += 1
            # agent is on a tile where a dropped bottle lies and it has not reached the carrying limit
            elif self.location in range(1, self.size - 1) and self.bottles_dropped[self.location - 1] > 0 and self.bottles_carrying < 2 and self.unbreakable_bottles:
                # remove bottle from current tile
                self.bottles_dropped[self.location - 1] -= 1
                # add bottle to agent's inventory
                self.bottles_carrying += 1

        # next observation
        observation = self._get_obs()

        # calculate potential-based low impact measure
        #r2_t = phi(S_t) - phi(S_t-1)
        #sum_t(r2_t) = 0 -> no impact
        reward[2] = self.potential(observation) - old_potential

        info = {}
        return observation, reward, terminal, info
    
    def reset(self, seed=None, return_info=False, **kwargs):
        super().reset(seed=seed)
        self.np_random.seed(seed)
        self.r_star = 0
        self.location = self.size - 1
        self.bottles_carrying = 0
        self.bottles_delivered = 0
        self.bottles_dropped = [0]*(self.size - 2)
        state = self._get_obs()
        return (state, {}) if return_info else state
    
    def get_obs_idx(self, obs):
        multi_index = np.array([[obs["location"]],
                                [obs["bottles_carrying"]], 
                                [obs["bottles_delivered"]],
                                *[[bd > 0] for bd in obs["bottles_dropped"]]])
        return np.ravel_multi_index(multi_index, tuple([self.size, 3, 2, *([2]*(self.size - 2))]))
    
    def _get_obs(self):
        return {"location": self.location,
                "bottles_carrying": self.bottles_carrying,
                "bottles_delivered": self.bottles_delivered,
                "bottles_dropped": self.bottles_dropped.copy()}
    
    def render(self, mode="human"):
        if mode == 'rgb_array':
            return np.array([self.state[:3], *self.state[3]]) # return RGB frame suitable for video
        elif mode == 'human':
            print("-----")
            print(f"Location: {self.location}\nCarrying {self.bottles_carrying} bottles.\nDelivered {self.bottles_delivered} so far.\nBottles have been dropped at tiles {'1' if self.bottles_dropped[0] > 0 else ''} {'2' if self.bottles_dropped[1] > 0 else ''} {'3' if self.bottles_dropped[2] > 0 else ''}")
            print("-----")
        else:
            super(BreakableBottles, self).render(mode=mode) # just raise an exception

    def close(self):
        pass

    def seed(self, seed=None):
        self.seed = seed if not seed is None else np.random.randint(2**32) 

    def potential(self, obs):
        if sum(obs["bottles_dropped"]) > 0:
            return -1
        return 0

if __name__ == '__main__':
    from gym.spaces.utils import flatdim
    env = BreakableBottles(size=5, prob_drop=0.1)
    assert flatdim(env.action_space) == 3
    assert flatdim(env.observation_space) == 13

    done = False
    obs = env.reset()
    while True:
        env.render()
        obs, r, done, info = env.step(env.action_space.sample())
        if done:
            break


from pathlib import Path

import gym
import numpy as np
import pygame
from gym.spaces import Box, Discrete


# As in Yang et al. (2019):
DEFAULT_MAP = np.array(
            [[0,    0,    0,   0,   0,  0,   0,   0,   0,   0,   0],
             [0.7,  0,    0,   0,   0,  0,   0,   0,   0,   0,   0],
             [-10,  8.2,  0,   0,   0,  0,   0,   0,   0,   0,   0],
             [-10, -10, 11.5,  0,   0,  0,   0,   0,   0,   0,   0],
             [-10, -10, -10, 14.0, 15.1,16.1,0,   0,   0,   0,   0],
             [-10, -10, -10, -10, -10, -10,  0,   0,   0,   0,   0],
             [-10, -10, -10, -10, -10, -10,  0,   0,   0,   0,   0],
             [-10, -10, -10, -10, -10, -10, 19.6, 20.3,0,   0,   0],
             [-10, -10, -10, -10, -10, -10, -10, -10,  0,   0,   0],
             [-10, -10, -10, -10, -10, -10, -10, -10, 22.4, 0,   0],
             [-10, -10, -10, -10, -10, -10, -10, -10, -10, 23.7, 0]]
        )

# As in Vamplew et al. (2018):
CONCAVE_MAP = np.array(
            [[0,    0,    0,   0,   0,  0,   0,   0,   0,   0,   0],
             [1.0,  0,    0,   0,   0,  0,   0,   0,   0,   0,   0],
             [-10,  2.0,  0,   0,   0,  0,   0,   0,   0,   0,   0],
             [-10, -10,  3.0,  0,   0,  0,   0,   0,   0,   0,   0],
             [-10, -10, -10, 5.0,  8.0,16.0, 0 ,  0,   0,   0,   0],
             [-10, -10, -10, -10, -10, -10,  0,   0,   0,   0,   0],
             [-10, -10, -10, -10, -10, -10,  0,   0,   0,   0,   0],
             [-10, -10, -10, -10, -10, -10, 24.0, 50.0,0,   0,   0],
             [-10, -10, -10, -10, -10, -10, -10, -10,  0,   0,   0],
             [-10, -10, -10, -10, -10, -10, -10, -10, 74.0, 0,   0],
             [-10, -10, -10, -10, -10, -10, -10, -10, -10, 124.0,0]]
        )

class DeepSeaTreasure(gym.Env):
    """Deep Sea Treasure environment

    Adapted from: https://github.com/RunzheYang/MORL

    CCS weights: [1,0], [0.7,0.3], [0.67,0.33], [0.6,0.4], [0.56,0.44], [0.52,0.48], [0.5,0.5], [0.4,0.6], [0.3,0.7], [0, 1]
    """

    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 4}

    def __init__(self, dst_map=DEFAULT_MAP, float_state=False):
        self.size = 11
        self.window_size = 512
        self.window = None
        self.clock = None

        self.float_state = float_state

        # The map of the deep sea treasure (convex version)
        self.sea_map = dst_map
        assert self.sea_map.shape == DEFAULT_MAP.shape, "The map's shape must be 11x11"
        
        self.dir = {
            0: np.array([-1, 0], dtype=np.int32),  # up
            1: np.array([1, 0], dtype=np.int32),  # down
            2: np.array([0, -1], dtype=np.int32),  # left
            3: np.array([0, 1], dtype=np.int32)  # right
        }

        # state space specification: 2-dimensional discrete box
        obs_type = np.float32 if self.float_state else np.int32
        if self.float_state:
            self.observation_space = Box(low=0.0, high=1.0, shape=(2,), dtype=obs_type)
        else:
            self.observation_space = Box(low=0, high=10, shape=(2,), dtype=obs_type)

        # action space specification: 1 dimension, 0 up, 1 down, 2 left, 3 right
        self.action_space = Discrete(4)
        self.reward_space = Box(low=np.array([0, -1]), high=np.array([np.max(self.sea_map), -1]), dtype=np.float32)

        self.current_state = np.array([0, 0], dtype=np.int32)

    def get_map_value(self, pos):
        return self.sea_map[pos[0]][pos[1]]

    def is_valid_state(self, state):
        if state[0] >= 0 and state[0] <= 10 and state[1] >= 0 and state[1] <= 10:
            if self.get_map_value(state) != -10:
                return True
        return False
    
    def render(self, mode='human'):
        # The size of a single grid square in pixels
        pix_square_size = self.window_size / self.size
        if self.window is None:
            self.submarine_img = pygame.image.load(str(Path(__file__).parent.absolute()) + '/assets/submarine.png')
            self.submarine_img = pygame.transform.scale(self.submarine_img, (pix_square_size, pix_square_size))
            self.submarine_img = pygame.transform.flip(self.submarine_img, flip_x=True, flip_y=False)
            self.treasure_img = pygame.image.load(str(Path(__file__).parent.absolute()) + '/assets/treasure.png')
            self.treasure_img = pygame.transform.scale(self.treasure_img, (pix_square_size, pix_square_size))

        if self.window is None and mode == "human":
            pygame.init()
            pygame.display.init()
            self.window = pygame.display.set_mode((self.window_size, self.window_size))
        if self.clock is None and mode == "human":
            self.clock = pygame.time.Clock()

        self.font = pygame.font.SysFont(None, 30)
        canvas = pygame.Surface((self.window_size, self.window_size))
        canvas.fill((0, 105, 148))

        for i in range(self.sea_map.shape[0]):
            for j in range(self.sea_map.shape[1]):
                if self.sea_map[i,j] == -10:
                    pygame.draw.rect(
                        canvas,
                        (0, 0, 0),
                        pygame.Rect(
                            pix_square_size * np.array([j,i]) + 0.6,
                            (pix_square_size, pix_square_size),
                        ),
                    )
                elif self.sea_map[i,j] != 0:
                   canvas.blit(self.treasure_img, np.array([j,i]) * pix_square_size)
                   img = self.font.render(str(self.sea_map[i,j]), True, (255, 255, 255))
                   canvas.blit(img, np.array([j,i]) * pix_square_size + np.array([5, 20]))
 
        canvas.blit(self.submarine_img, self.current_state[::-1] * pix_square_size)

        for x in range(self.size + 1):
            pygame.draw.line(
                canvas,
                0,
                (0, pix_square_size * x),
                (self.window_size, pix_square_size * x),
                width=1,
            )
            pygame.draw.line(
                canvas,
                0,
                (pix_square_size * x, 0),
                (pix_square_size * x, self.window_size),
                width=1,
            )

        if mode == "human":
            # The following line copies our drawings from `canvas` to the visible window
            self.window.blit(canvas, canvas.get_rect())
            pygame.event.pump()
            pygame.display.update()

            # We need to ensure that human-rendering occurs at the predefined framerate.
            # The following line will automatically add a delay to keep the framerate stable.
            self.clock.tick(self.metadata["render_fps"])
        else:  # rgb_array
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(canvas)), axes=(1, 0, 2)
            )

    def get_state(self):
        if self.float_state:
            state = self.current_state.astype(np.float32) * 0.1
        else:
            state = self.current_state.copy()
        return state

    def reset(self, seed=None, return_info=False, **kwargs):
        super().reset(seed=seed)
        self.np_random.seed(seed)

        self.current_state = np.array([0, 0], dtype=np.int32)
        self.step_count = 0.0
        state = self.get_state()
        return (state, {}) if return_info else state

    def step(self, action):
        next_state = self.current_state + self.dir[action]

        if self.is_valid_state(next_state):
            self.current_state = next_state

        treasure_value = self.get_map_value(self.current_state)
        if treasure_value == 0 or treasure_value == -10:
            treasure_value = 0.0
            terminal = False
        else:
            terminal = True
        time_penalty = -1.0
        vec_reward = np.array([treasure_value, time_penalty], dtype=np.float32)

        state = self.get_state()

        return state, vec_reward, terminal, {}

    def close(self):
        if self.window is not None:
            pygame.display.quit()
            pygame.quit()


if __name__ == '__main__':

    env = DeepSeaTreasure()
    done = False
    env.reset()
    while True:
        env.render()
        obs, r, done, info = env.step(env.action_space.sample())
        if done:
            env.reset()



import mo_gym.deep_sea_treasure
import mo_gym.four_room
import mo_gym.minecart
import mo_gym.mountain_car
import mo_gym.resource_gathering
import mo_gym.fruit_tree
import mo_gym.breakable_bottles
import mo_gym.mario
import mo_gym.reacher
import mo_gym.mujoco
from mo_gym.utils import *
from mo_gym.evaluation import *



import gym
from gym import spaces
import numpy as np
from pybulletgym.envs.roboschool.robots.robot_bases import MJCFBasedRobot
from pybulletgym.envs.roboschool.envs.env_bases import BaseBulletEnv
from pybulletgym.envs.roboschool.scenes.scene_bases import SingleRobotEmptyScene

target_positions = list(map(lambda l: np.array(l), [(0.14, 0.0), (-0.14, 0.0), (0.0, 0.14), (0.0, -0.14)]))


class ReacherBulletEnv(BaseBulletEnv):

    def __init__(self, target=(0.14, 0.0), fixed_initial_state=False):
        self.robot = ReacherRobot(target, fixed_initial_state=fixed_initial_state)
        BaseBulletEnv.__init__(self, self.robot)
        self._cam_dist = 0.75

        #self.target_positions = list(map(lambda l: np.array(l), [(0.14, 0.0), (-0.14, 0.0), (0.0, 0.14), (0.0, -0.14), (0.22, 0.0), (-0.22, 0.0), (0.0, 0.22), (0.0, -0.22), (0.1, 0.1), (0.1, -0.1), (-0.1, 0.1), (-0.1, -0.1)]))
        #self.target_positions = list(map(lambda l: np.array(l), [(0.14, 0.0), (-0.14, 0.0), (0.0, 0.14), (0.0, -0.14), (0.1, 0.1), (0.1, -0.1), (-0.1, 0.1), (-0.1, -0.1)]))
        self.target_positions = list(map(lambda l: np.array(l), [(0.14, 0.0), (-0.14, 0.0), (0.0, 0.14), (0.0, -0.14)]))

        actions = [-1., 0., 1.]
        self.action_dict = dict()
        for a1 in actions:
            for a2 in actions:
                self.action_dict[len(self.action_dict)] = (a1, a2)
        
        self.action_space = spaces.Discrete(9)
        self.observation_space = spaces.Box(low=-np.inf, high=np.inf, shape=(5,), dtype=np.float32)
        self.reward_space = spaces.Box(low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32)

    def create_single_player_scene(self, bullet_client):
        return SingleRobotEmptyScene(bullet_client, gravity=0.0, timestep=0.0165, frame_skip=1)

    def step(self, a):
        real_action = self.action_dict[a]

        assert (not self.scene.multiplayer)
        self.robot.apply_action(real_action)
        self.scene.global_step()

        state = self.robot.calc_state()  # sets self.to_target_vec
        
        """ delta = np.linalg.norm(np.array(self.robot.fingertip.pose().xyz()) - np.array(self.robot.target.pose().xyz()))
        reward = 1. - 4. * delta """

        phi = np.zeros(len(self.target_positions))
        for index, target in enumerate(self.target_positions):
            delta = np.linalg.norm(np.array(self.robot.fingertip.pose().xyz()[:2]) - target)
            phi[index] = (1. - 4*delta) # 1 - 4

        self.HUD(state, real_action, False)
        
        return state, phi, False, {}

    def camera_adjust(self):
        x, y, z = self.robot.fingertip.pose().xyz()
        x *= 0.5
        y *= 0.5
        self.camera.move_and_look_at(0.3, 0.3, 0.3, x, y, z)
    
    def reset(self, seed=None, return_info=False, **kwargs):
        self.np_random.seed(seed)
        return (super().reset(), {}) if return_info else super().reset()


class ReacherRobot(MJCFBasedRobot):
    TARG_LIMIT = 0.27

    def __init__(self, target, fixed_initial_state=False):
        MJCFBasedRobot.__init__(self, 'reacher.xml', 'body0', action_dim=2, obs_dim=4)
        self.target_pos = target
        self.fixed_initial_state = fixed_initial_state

    def robot_specific_reset(self, bullet_client):
        self.jdict["target_x"].reset_current_position(target_positions[0][0], 0)
        self.jdict["target_y"].reset_current_position(target_positions[0][1], 0)

        """ self.jdict["target2_x"].reset_current_position(target_positions[1][0], 0)
        self.jdict["target2_y"].reset_current_position(target_positions[1][1], 0)
        self.jdict["target3_x"].reset_current_position(target_positions[2][0], 0)
        self.jdict["target3_y"].reset_current_position(target_positions[2][1], 0)
        self.jdict["target4_x"].reset_current_position(target_positions[3][0], 0)
        self.jdict["target4_y"].reset_current_position(target_positions[3][1], 0) """

        self.fingertip = self.parts["fingertip"]
        self.target = self.parts["target"]
        self.central_joint = self.jdict["joint0"]
        self.elbow_joint = self.jdict["joint1"]
        if not self.fixed_initial_state:
            self.central_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
            self.elbow_joint.reset_current_position(self.np_random.uniform(low=-3.14 / 2, high=3.14 / 2), 0)
        else:
            self.central_joint.reset_current_position(0, 0)
            self.elbow_joint.reset_current_position(3.1415/2, 0)

    def apply_action(self, a):
        assert (np.isfinite(a).all())
        self.central_joint.set_motor_torque(0.05 * float(np.clip(a[0], -1, +1)))
        self.elbow_joint.set_motor_torque(0.05 * float(np.clip(a[1], -1, +1)))

    def calc_state(self):
        theta, self.theta_dot = self.central_joint.current_relative_position()
        self.gamma, self.gamma_dot = self.elbow_joint.current_relative_position()
        # target_x, _ = self.jdict["target_x"].current_position()
        # target_y, _ = self.jdict["target_y"].current_position()
        self.to_target_vec = np.array(self.fingertip.pose().xyz()) - np.array(self.target.pose().xyz())
        return np.array([
            np.cos(theta),
            np.sin(theta),
            self.theta_dot*0.1,
            self.gamma,
            self.gamma_dot*0.1,
        ], dtype=np.float32)


if __name__ == '__main__':

    env = ReacherBulletEnv()
    #env.render(mode='human')
    obs = env.reset()
    print(env.observation_space.contains(obs), obs.dtype, env.observation_space)
    while True:
        env.step(env.action_space.sample())
        #env.render(mode='human')


from copy import deepcopy
from typing import Tuple, TypeVar, Iterator

import gym
import numpy as np
import time
from gym.vector import SyncVectorEnv
from gym.wrappers.normalize import RunningMeanStd
from gym.wrappers import RecordEpisodeStatistics

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


def make(env_name: str, disable_env_checker: bool = True, **kwargs) -> gym.Env:
    """ Disable env checker, as it requires the reward to be a scalar."""
    return gym.make(env_name, disable_env_checker=disable_env_checker, **kwargs)


class LinearReward(gym.Wrapper):
    """Wrapper for Multi-Objective Envs
    Makes the env return a scalar reward, which is the the dot-product between the reward vector and the weight vector.
    """

    def __init__(self, env: gym.Env, weight: np.ndarray = None):
        super().__init__(env)
        if weight is None:
            weight = np.ones(shape=env.reward_space.shape)
        self.set_weight(weight)

    def set_weight(self, weight):
        assert weight.shape == self.env.reward_space.shape, "Reward weight has different shape than reward vector."
        self.w = weight

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, dict]:
        observation, reward, done, info = self.env.step(action)
        scalar_reward = np.dot(reward, self.w)
        info['vector_reward'] = reward

        return observation, scalar_reward, done, info


class MONormalizeReward(gym.Wrapper):
    """
    Wrapper to normalize the reward component at index idx. Does not touch other reward components.
    Based on Gym's implementation: https://github.com/openai/gym/blob/master/gym/wrappers/normalize.py#L113
    """

    def __init__(self, env: gym.Env, idx: int, gamma: float = 0.99, epsilon: float = 1e-8):
        """This wrapper will normalize immediate rewards s.t. their exponential moving average has a fixed variance.

             Args:
                 env (env): The environment to apply the wrapper
                 idx (int): the index of the reward to normalize
                 epsilon (float): A stability parameter
                 gamma (float): The discount factor that is used in the exponential moving average.
             """
        super().__init__(env)
        self.idx = idx
        self.num_envs = getattr(env, "num_envs", 1)
        self.is_vector_env = getattr(env, "is_vector_env", False)
        self.return_rms = RunningMeanStd(shape=())
        self.returns = np.zeros(self.num_envs)
        self.gamma = gamma
        self.epsilon = epsilon

    def step(self, action: ActType):
        """Steps through the environment, normalizing the rewards returned."""
        obs, rews, dones, infos = self.env.step(action)
        to_normalize = rews[self.idx]
        if not self.is_vector_env:
            to_normalize = np.array([to_normalize])
        self.returns = self.returns * self.gamma + to_normalize
        to_normalize = self.normalize(to_normalize)
        self.returns[dones] = 0.0
        if not self.is_vector_env:
            to_normalize = to_normalize[0]
        rews[self.idx] = to_normalize
        return obs, rews, dones, infos

    def normalize(self, rews):
        """Normalizes the rewards with the running mean rewards and their variance."""
        self.return_rms.update(self.returns)
        return rews / np.sqrt(self.return_rms.var + self.epsilon)


class MOClipReward(gym.RewardWrapper):
    """"Clip reward[idx] to [min, max]. """

    def __init__(self, env: gym.Env, idx: int, min_r, max_r):
        super().__init__(env)
        self.idx = idx
        self.min_r = min_r
        self.max_r = max_r

    def reward(self, reward):
        reward[self.idx] = np.clip(reward[self.idx], self.min_r, self.max_r)
        return reward


class MOSyncVectorEnv(SyncVectorEnv):
    """Vectorized environment that serially runs multiple environments.
    """

    def __init__(
            self,
            env_fns: Iterator[callable],
            copy: bool = True,
    ):
        super().__init__(
            env_fns,
            copy=copy
        )
        # Just overrides the rewards memory to add the number of objectives
        self.reward_space = self.envs[0].reward_space
        self._rewards = np.zeros((self.num_envs, self.reward_space.shape[0],), dtype=np.float64)

def add_vector_episode_statistics(
    info: dict, episode_info: dict, num_envs: int, num_objs: int, env_num: int
):
    """Add episode statistics.

    Add statistics coming from the vectorized environment.

    Args:
        info (dict): info dict of the environment.
        episode_info (dict): episode statistics data.
        num_envs (int): number of environments.
        num_envs (int): number of objectives.
        env_num (int): env number of the vectorized environments.

    Returns:
        info (dict): the input info dict with the episode statistics.
    """
    info["episode"] = info.get("episode", {})

    info["_episode"] = info.get("_episode", np.zeros(num_envs, dtype=bool))
    info["_episode"][env_num] = True

    for k in episode_info.keys():
        if k == "r" or k == "dr":
            info_array = info["episode"].get(k, np.zeros((num_envs, num_objs)))
        else:
            info_array = info["episode"].get(k, np.zeros(num_envs))
        info_array[env_num] = deepcopy(episode_info[k])
        info["episode"][k] = info_array

    return info

class MORecordEpisodeStatistics(RecordEpisodeStatistics):
    def __init__(self, env: gym.Env, gamma: float = 1., deque_size: int = 100):
        """This wrapper will keep track of cumulative rewards and episode lengths.

        Args:
            env (Env): The environment to apply the wrapper
            deque_size: The size of the buffers :attr:`return_queue` and :attr:`length_queue`
        """
        super().__init__(env, deque_size)
        # Here we just override the standard implementation to extend to MO
        self.reward_dim = self.env.reward_space.shape[0]
        self.gamma = gamma

    def reset(self, **kwargs):
        """Resets the environment using kwargs and resets the episode returns and lengths."""
        observations = super().reset(**kwargs)
        self.episode_returns = np.zeros((self.num_envs, self.reward_dim), dtype=np.float32)
        self.disc_episode_returns = np.zeros((self.num_envs, self.reward_dim), dtype=np.float32)
        return observations

    def step(self, action):
        """Steps through the environment, recording the episode statistics."""
        # This is the code from the RecordEpisodeStatistics wrapper from gym.
        observations, rewards, dones, infos = self.env.step(action)
        assert isinstance(
            infos, dict
        ), f"`info` dtype is {type(infos)} while supported dtype is `dict`. This may be due to usage of other wrappers in the wrong order."
        self.episode_returns += rewards
        self.disc_episode_returns += (rewards * np.repeat(self.gamma ** self.episode_lengths, self.reward_dim).reshape(self.episode_returns.shape))
        self.episode_lengths += 1
        if not self.is_vector_env:
            dones = [dones]
        dones = list(dones)

        for i in range(len(dones)):
            if dones[i]:
                episode_return = deepcopy(self.episode_returns[i])  # Makes a deepcopy to avoid subsequent mutations
                disc_episode_return = deepcopy(self.disc_episode_returns[i])  # Makes a deepcopy to avoid subsequent mutations
                episode_length = self.episode_lengths[i]
                episode_info = {
                    "episode": {
                        "r": episode_return,
                        "dr": disc_episode_return,
                        "l": episode_length,
                        "t": round(time.perf_counter() - self.t0, 6),
                    }
                }
                if self.is_vector_env:
                    infos = add_vector_episode_statistics(
                        infos, episode_info["episode"], self.num_envs, self.reward_dim, i
                    )
                else:
                    infos = {**infos, **episode_info}
                self.return_queue.append(episode_return)
                self.length_queue.append(episode_length)
                self.episode_count += 1
                self.episode_returns[i] = 0
                self.disc_episode_returns[i] = 0
                self.episode_lengths[i] = 0
        return (
            observations,
            rewards,
            dones if self.is_vector_env else dones[0],
            infos,
        )




from pathlib import Path

from gym.envs.registration import register

register(
    id='minecart-v0',
    entry_point='mo_gym.minecart.minecart:Minecart',
    max_episode_steps=1000
)

register(
    id='minecart-deterministic-v0',
    entry_point='mo_gym.minecart.minecart:Minecart',
    kwargs={'config': str(Path(__file__).parent.absolute()) + '/mine_config_det.json'},
    max_episode_steps=1000
)



from typing import Tuple, List, Union

from pymoo.indicators.hv import HV
import numpy as np


def hypervolume(ref_point: np.ndarray, points: List[np.ndarray]) -> float:
    """Computes the hypervolume metric for a set of points (value vectors) and a reference point.

    Args:
        ref_point (np.ndarray): Reference point
        points (List[np.ndarray]): List of value vectors

    Returns:
        float: Hypervolume metric
    """    
    return HV(ref_point=ref_point * - 1)(np.array(points) * - 1)


def eval_mo(agent, env, w: np.ndarray, render: bool = False) -> Tuple[float, float, np.ndarray, np.ndarray]:
    """Evaluates one episode of the agent in the environment.

    Args:
        agent: Agent
        env: MO-Gym environment with LinearReward wrapper
        w (np.ndarray): Weight vector
        render (bool, optional): Whether to render the environment. Defaults to False.

    Returns:
        (float, float, np.ndarray, np.ndarray): Scalarized total reward, scalarized return, vectorized total reward, vectorized return
    """    
    obs = env.reset()
    done = False
    total_vec_reward, vec_return = np.zeros_like(w), np.zeros_like(w)
    gamma = 1.0
    while not done:
        if render:
            env.render(mode='human')
        obs, r, done, info = env.step(agent.eval(obs, w))
        total_vec_reward += r
        vec_return += gamma * r
        gamma *= agent.gamma
    return (
        np.dot(w, total_vec_reward),
        np.dot(w, vec_return),
        total_vec_reward,
        vec_return,
    )


def policy_evaluation_mo(agent, env, w: np.ndarray, rep: int = 5, return_scalarized_value: bool = False) -> Union[np.ndarray, float]:
    """Evaluates the value of a policy by runnning the policy for multiple episodes.

    Args:
        agent: Agent
        env: MO-Gym environment with LinearReward wrapper
        w (np.ndarray): Weight vector
        rep (int, optional): Number of episodes for averaging. Defaults to 5.
        return_scalarized_value (bool, optional): Whether to return scalarized value. Defaults to False.

    Returns:
        np.ndarray: Value of the policy
    """    
    if return_scalarized_value:
        returns = [eval_mo(agent, env, w)[1] for _ in range(rep)]
    else:
        returns = [eval_mo(agent, env, w)[3] for _ in range(rep)]
    return np.mean(returns, axis=0)


from pathlib import Path

import gym
import numpy as np
import pygame
from gym.spaces import Box, Discrete


class ResourceGathering(gym.Env):
    """Resource Gathering environment

    From "Barrett, Leon & Narayanan, Srini. (2008). Learning all optimal policies with multiple criteria. 
    Proceedings of the 25th International Conference on Machine Learning. 41-47. 10.1145/1390156.1390162."
    """

    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 4}

    def __init__(self):
        self.size = 5
        self.window_size = 512
        self.window = None
        self.clock = None

        # The map of resource gathering env
        self.map = np.array([
            [' ', ' ', 'R1', 'E2', ' '],
            [' ', ' ', 'E1', ' ', 'R2'],
            [' ', ' ', ' ', ' ', ' '],
            [' ', ' ', ' ', ' ', ' '],
            [' ', ' ', 'H', ' ', ' ']
            ]
        )
        self.initial_pos = np.array([4, 2], dtype=np.int32)

        self.dir = {
            0: np.array([-1, 0], dtype=np.int32),  # up
            1: np.array([1, 0], dtype=np.int32),  # down
            2: np.array([0, -1], dtype=np.int32),  # left
            3: np.array([0, 1], dtype=np.int32)  # right
        }

        self.observation_space = Box(low=0.0, high=5.0, shape=(4,), dtype=np.int32)

        # action space specification: 1 dimension, 0 up, 1 down, 2 left, 3 right
        self.action_space = Discrete(4)
        # reward space: 
        self.reward_space = Box(low=-1, high=1, shape=(3,), dtype=np.float32)

    def get_map_value(self, pos):
        return self.map[pos[0]][pos[1]]

    def is_valid_state(self, state):
        return state[0] >= 0 and state[0] < self.size and state[1] >= 0 and state[1] < self.size
    
    def render(self, mode='human'):
        # The size of a single grid square in pixels
        pix_square_size = self.window_size / self.size
        if self.window is None:
            self.gold_img = pygame.image.load(str(Path(__file__).parent.absolute()) + '/assets/gold.png')
            self.gold_img = pygame.transform.scale(self.gold_img, (pix_square_size, pix_square_size))
            self.gem_img = pygame.image.load(str(Path(__file__).parent.absolute()) + '/assets/gem.png')
            self.gem_img = pygame.transform.scale(self.gem_img, (pix_square_size, pix_square_size))
            self.enemy_img = pygame.image.load(str(Path(__file__).parent.absolute()) + '/assets/sword.png')
            self.enemy_img = pygame.transform.scale(self.enemy_img, (pix_square_size, pix_square_size))
            self.home_img = pygame.image.load(str(Path(__file__).parent.absolute()) + '/assets/home.png')
            self.home_img = pygame.transform.scale(self.home_img, (pix_square_size, pix_square_size))
            self.agent_img = pygame.image.load(str(Path(__file__).parent.absolute()) + '/assets/stickerman.png')
            self.agent_img = pygame.transform.scale(self.agent_img, (pix_square_size, pix_square_size))

        if self.window is None and mode == "human":
            pygame.init()
            pygame.display.init()
            self.window = pygame.display.set_mode((self.window_size, self.window_size))
        if self.clock is None and mode == "human":
            self.clock = pygame.time.Clock()

        canvas = pygame.Surface((self.window_size, self.window_size))
        canvas.fill((255, 255, 255))

        canvas.blit(self.home_img, self.initial_pos[::-1] * pix_square_size)
        for i in range(self.map.shape[0]):
            for j in range(self.map.shape[1]):
                pos = np.array([j,i])
                if self.map[i,j] == 'R1' and not self.has_gold:
                    canvas.blit(self.gold_img, np.array([j,i]) * pix_square_size)
                elif self.map[i,j] == 'R2' and not self.has_gem:
                    canvas.blit(self.gem_img, np.array([j,i]) * pix_square_size)
                elif self.map[i,j] == 'E1' or self.map[i,j] == 'E2':
                    canvas.blit(self.enemy_img, np.array([j,i]) * pix_square_size)
            
        canvas.blit(self.agent_img, self.current_pos[::-1] * pix_square_size)

        for x in range(self.size + 1):
            pygame.draw.line(
                canvas,
                0,
                (0, pix_square_size * x),
                (self.window_size, pix_square_size * x),
                width=2,
            )
            pygame.draw.line(
                canvas,
                0,
                (pix_square_size * x, 0),
                (pix_square_size * x, self.window_size),
                width=2,
            )

        if mode == "human":
            # The following line copies our drawings from `canvas` to the visible window
            self.window.blit(canvas, canvas.get_rect())
            pygame.event.pump()
            pygame.display.update()

            # We need to ensure that human-rendering occurs at the predefined framerate.
            # The following line will automatically add a delay to keep the framerate stable.
            self.clock.tick(self.metadata["render_fps"])
        else:  # rgb_array
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(canvas)), axes=(1, 0, 2)
            )

    def get_state(self):
        pos = self.current_pos.copy()
        state = np.concatenate((pos, np.array([self.has_gold, self.has_gem], dtype=np.int32)))
        return state

    def reset(self, seed=None, return_info=False, **kwargs):
        super().reset(seed=seed)
        self.np_random.seed(seed)

        self.current_pos = self.initial_pos
        self.has_gem = 0
        self.has_gold = 0
        self.step_count = 0.0
        state = self.get_state()
        return (state, {}) if return_info else state

    def step(self, action):
        next_pos = self.current_pos + self.dir[action]

        if self.is_valid_state(next_pos):
            self.current_pos = next_pos

        vec_reward = np.zeros(3, dtype=np.float32)
        done = False

        cell = self.get_map_value(self.current_pos)
        if cell == 'R1':
            self.has_gold = 1
        elif cell == 'R2':
            self.has_gem = 1
        elif cell == 'E1' or cell == 'E2':
            if np.random.random() < 0.1:
                vec_reward[0] = -1.0
                done = True
        elif cell == 'H':
            done = True
            vec_reward[1] = self.has_gold
            vec_reward[2] = self.has_gem

        state = self.get_state()

        return state, vec_reward, done, {}

    def close(self):
        if self.window is not None:
            pygame.display.quit()
            pygame.quit()


if __name__ == '__main__':

    env = ResourceGathering()
    done = False
    env.reset()
    while True:
        env.render()
        obs, r, done, info = env.step(env.action_space.sample())
        print(obs, r, done)
        if done:
            env.reset()



import numpy as np
import gym
#from stable_baselines3.common.atari_wrappers import MaxAndSkipEnv
from gym.wrappers import (FrameStack, GrayScaleObservation, ResizeObservation, TimeLimit)
from gym.utils import seeding
from nes_py.wrappers import JoypadSpace
from gym_super_mario_bros.actions import SIMPLE_MOVEMENT
from gym_super_mario_bros import SuperMarioBrosEnv
import gym_super_mario_bros
#import matplotlib.pyplot as plt

import mo_gym


class MOSuperMarioBros(SuperMarioBrosEnv):
    
    def __init__(self, rom_mode='pixel', lost_levels=False, target=None, objectives=['x_pos', 'time', 'death', 'coin', 'enemy']):
        super().__init__(rom_mode, lost_levels, target)

        self.objectives = set(objectives)
        self.reward_space = gym.spaces.Box(high=np.inf, low=-np.inf, shape=(len(objectives),))

        self.single_stage = True
        self.done_when_dead = True

    def reset(self, seed=None, return_info=False, **kwargs):
        self._np_random, seed = seeding.np_random(seed) # this is not used
        self.coin = 0
        self.x_pos = 0
        self.time = 0
        self.score = 0
        self.stage_bonus = 0
        self.lives = 2
        return (super().reset(), {}) if return_info else super().reset()

    def step(self, action):
        obs, reward, done, info = super().step(action)

        if self.single_stage and info["flag_get"]:
            self.stage_bonus = 10000
            done = True

        ''' Construct Multi-Objective Reward'''
        # [x_pos, time, death, coin, enemy]
        moreward = []

        # 1. x position
        if 'x_pos' in self.objectives:
            xpos_r = info["x_pos"] - self.x_pos
            self.x_pos = info["x_pos"]
            # resolve an issue where after death the x position resets
            if xpos_r < -5:
                xpos_r = 0
            moreward.append(xpos_r)
        
        # 2. time penaltiy 
        if 'time' in self.objectives:
            time_r = info["time"] - self.time
            self.time = info["time"]
            # time is aways decreasing
            if time_r > 0:
                time_r = 0
            moreward.append(time_r)

        # 3. death
        if 'death' in self.objectives:
            if self.lives > info['life']:
                death_r = -25
            else:
                death_r = 0
            moreward.append(death_r)

        # 4. coin
        if 'coin' in self.objectives:
            coin_r = (info['coins'] - self.coin) * 100
            self.coin = info['coins']
            moreward.append(coin_r)

        # 5. enemy
        if 'enemy' in self.objectives:
            enemy_r = info['score'] - self.score
            if coin_r > 0 or done:
                enemy_r = 0
            self.score = info['score']
            moreward.append(enemy_r)

        ############################################################################

        if self.done_when_dead:
            # when Mario loses life, changes the state to the terminal
            if self.lives > info['life'] and info['life'] > 0:
                done = True

        self.lives = info['life']
       
        mor = np.array(moreward, dtype=np.float32) * self.reward_space.shape[0] / 150

        info['score'] = info['score'] + self.stage_bonus

        return obs, mor, bool(done), info


if __name__ == '__main__':

    env = MOSuperMarioBros()
    env = JoypadSpace(env, SIMPLE_MOVEMENT)
    #env = MaxAndSkipEnv(env, 4)
    env = ResizeObservation(env, (84, 84))
    env = GrayScaleObservation(env)
    #env = FrameStack(env, 4)
    env = mo_gym.LinearReward(env)

    done = False
    env.reset()
    while True:
        obs, r, done, info = env.step(env.action_space.sample())
        print(r, info['vector_reward'], done, info['time'])
        """ plt.figure()
        plt.imshow(obs, cmap='gray', vmin=0, vmax=255)
        plt.show() """
        env.render()
        if done:
            env.reset()



from setuptools import setup, find_packages

# Environment-specific dependencies.
extras = {
    "mario": ["nes-py", "gym-super-mario-bros"],
    "minecart" : ["scipy"]
}
extras["all"] = extras["mario"] + extras["minecart"]

packages = [package for package in find_packages() if package.startswith("mo_gym")]
package_data = {package: ["*.json", "assets/*"] for package in packages}

setup(
    name="mo-gym",
    version="0.1.2",
    description="Environments for Multi-Objective RL.",
    url="https://www.github.com/LucasAlegre/mo-gym",
    author="LucasAlegre",
    author_email="lnalegre@inf.ufrgs.br",
    license="MIT",
    packages=packages,
    package_data=package_data,
    install_requires=[
        "gym==0.24.1", # 0.25 has breaking changes
        "numpy",
        "pygame",
        "scipy",
        "pymoo",
    ],
    extras_require=extras,
    tests_require=["pytest", "mock"],
    python_requires=">=3.7",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)

# https://towardsdatascience.com/create-your-own-python-package-and-publish-it-into-pypi-9306a29bc116


from gym.envs.registration import register


register(
    id='mo-halfcheetah-v4',
    entry_point='mo_gym.mujoco.half_cheetah:MOHalfCheehtahEnv',
    max_episode_steps=1000,
)

register(
    id='mo-hopper-v4',
    entry_point='mo_gym.mujoco.hopper:MOHopperEnv',
    max_episode_steps=1000,
)